/************************************************************************
 *
 *  gethdr.c, header decoding for tmndecode (H.263 decoder)
 *  Copyright (C) 1995, 1996  Telenor R&D, Norway
 *
 *  Contacts:
 *  Robert Danielsen                  <Robert.Danielsen@nta.no>
 *
 *  Telenor Research and Development  http://www.nta.no/brukere/DVC/
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *
 *  Copyright (C) 1997  University of BC, Canada
 *  Modified by: Michael Gallant <mikeg@ee.ubc.ca>
 *               Guy Cote <guyc@ee.ubc.ca>
 *               Berna Erol <bernae@ee.ubc.ca>
 *
 *  Contacts:
 *  Michael Gallant                   <mikeg@ee.ubc.ca>
 *
 *  UBC Image Processing Laboratory   http://www.ee.ubc.ca/image
 *  2356 Main Mall                    tel.: +1 604 822 4051
 *  Vancouver BC Canada V6T1Z4        fax.: +1 604 822 5949
 *
 ************************************************************************/

/* Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee
 * or royalty on an "as is" basis. The University of British Columbia
 * disclaims any and all warranties, whether express, implied, or
 * statuary, including any implied warranties or merchantability or of
 * fitness for a particular purpose.  In no event shall the
 * copyright-holder be liable for any incidental, punitive, or
 * consequential damages of any kind whatsoever arising from the use of
 * these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and
 * user's customers, employees, agents, transferees, successors, and
 * assigns.
 *
 * The University of British Columbia does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any
 * third-party patents.
 *
 * Commercial implementations of H.263, including shareware, are subject to
 * royalty fees to patent holders.  Many of these patents are general
 * enough such that they are unavoidable regardless of implementation
 * design.
 *
 */


/* modified to support annex O true B frames, mikeg@ee.ubc.ca
 *
 * modified to support general H.263+ syntax, bernae@ee.ubc.ca
 *
 * based on mpeg2decode, (C) 1994, MPEG Software Simulation Group and
 * mpeg2play, (C) 1994 Stefan Eckart <stefan@lis.e-technik.tu-muenchen.de> */


#include <stdio.h>
#include <stdlib.h>

#include "tmndec.h"
#include "global.h"
#include "../error.h"

/************ Prototypes ************/

int read_H263_getheader();
void read_H263_startcode();
static void read_H263_getpicturehdr();
void read_H263_getgobheader();
void read_H263_initoptions();

/************************************/


/* decode headers from one input stream until an End of Sequence or
 * picture start code is found */

int read_H263_getheader()
{
	unsigned int code, gob;
	extern int read_H263_picnr;

	/* look for startcode */
	read_H263_startcode();
	code = read_H263_getbits(PSC_LENGTH);
	gob = read_H263_getbits(5);
	if (gob == SE_CODE) return 0;
	if (gob == 0) {
		read_H263_picnr = read_H263_getbits(16)<<16;
		read_H263_picnr |= read_H263_getbits(16);
		read_H263_getpicturehdr();
		if (read_H263_syntax_arith_coding) read_H263_decoder_reset();             /* reset decoder after receiving fixed length PSC string */
	}
	return gob + 1;
}

/* align to start of next startcode */
void read_H263_startcode()
{
	/* search for new picture start code */
	while (read_H263_showbits(PSC_LENGTH) != 1l) {
		read_H263_flushbits(1);
	}
}

/* decode picture header */
static void read_H263_getpicturehdr()
{
	int pos, pei, tmp;
	int TRPI = 0;
	int BCI = 0;
	int clock_conversion_code = 0;
	int clock_divisor = 0;
	int extended_temporal_reference = 0;

	read_H263_UFEP = 0;

	pos = read_H263_ld->bitcnt;

	read_H263_temp_ref = read_H263_getbits(8);
	if (read_H263_trd < 0) read_H263_trd += 256;

	tmp = read_H263_getbits(1);            /* always "1" */
	if (!tmp && !read_H263_quiet) printf("warning: spare in picture header should be \"1\"\n");
	tmp = read_H263_getbits(1);            /* always "0" */
	if (tmp && !read_H263_quiet) printf("warning: H.261 distinction bit should be \"0\"\n");
	tmp = read_H263_getbits(1);            /* split_screen_indicator */
	ASSERT(!tmp,(MSG("split-screen not supported in this version")));
	tmp = read_H263_getbits(1);            /* document_camera_indicator */
	if (tmp && !read_H263_quiet) printf("warning: document camera indicator not supported in this version\n");
	tmp = read_H263_getbits(1);            /* freeze_picture_release */
	if (tmp && !read_H263_quiet) printf ("warning: frozen picture not supported in this version\n");
	tmp = read_H263_getbits(3);
	ASSERT(tmp,(MSG("source format 000 is reserved and not used in this version")));
	if (tmp == EXTENDED_PTYPE) {
		read_H263_plus_type = 1;
		read_H263_UFEP = read_H263_getbits(3);
		printf("UFEP: %i\n",read_H263_UFEP);
		if (read_H263_UFEP == 1) {                           /* OPPTYPE */
			read_H263_source_format = read_H263_getbits(3);
			/* optional custom picture clock frequency */
			read_H263_optional_custom_PCF = read_H263_getbits(1);
			ASSERT(!read_H263_optional_custom_PCF,(MSG("Optional custom picture clock frequency is not supported in this version")));
			read_H263_mv_outside_frame = read_H263_getbits(1);
			read_H263_long_vectors = (read_H263_mv_outside_frame ? 1 : 0);
			read_H263_syntax_arith_coding = read_H263_getbits(1);
			read_H263_adv_pred_mode = read_H263_getbits(1);
			read_H263_mv_outside_frame = (read_H263_adv_pred_mode ? 1 : read_H263_mv_outside_frame);
			read_H263_overlapping_MC = (read_H263_adv_pred_mode ? 1 : 0);
			read_H263_use_4mv = (read_H263_adv_pred_mode ? 1 : 0);
			read_H263_pb_frame = 0;
			read_H263_advanced_intra_coding = read_H263_getbits(1);
			read_H263_deblocking_filter_mode = read_H263_getbits(1);
			read_H263_mv_outside_frame = (read_H263_deblocking_filter_mode ? 1 : read_H263_mv_outside_frame);
			read_H263_use_4mv = (read_H263_deblocking_filter_mode ? 1 : read_H263_use_4mv);
			read_H263_slice_structured_mode = read_H263_getbits(1);
			ASSERT(!read_H263_slice_structured_mode,(MSG("Slice structured mode is not supported in this version")));
			read_H263_reference_picture_selection_mode = read_H263_getbits(1);
			read_H263_independently_segmented_decoding_mode = read_H263_getbits(1);
			ASSERT(!read_H263_independently_segmented_decoding_mode,(MSG("Independently segmented decoding mode is not supported in this version")));
			read_H263_alternative_inter_VLC_mode = read_H263_getbits(1);
			read_H263_modified_quantization_mode = read_H263_getbits(1);
			tmp = read_H263_getbits(4);
			ASSERT(tmp == 8,(MSG("The last 4 bits of OPPTYPE is expected to be 1000"))); /* OPPTYPE : bit15=1, bit16,bit17,bit18=0 */
		}
		if ((read_H263_UFEP == 1) || (read_H263_UFEP == 0)) {
			if (read_H263_UFEP == 0) {
				if (read_H263_scalability_mode >= 3) {
					read_H263_horizontal_size = read_H263_lines[read_H263_base_source_format];
					read_H263_vertical_size = read_H263_pels[read_H263_base_source_format];

					read_H263_mb_width = read_H263_horizontal_size / 16;
					read_H263_mb_height = read_H263_vertical_size / 16;

					/* Need to store previous (reference layer) values
					 * for interpolation purposes, as the new values,
					 * i.e. of the spatially scaled layer, depend on
					 * the type of spatial scalability in use. */
					read_H263_ref_coded_picture_width = read_H263_coded_picture_width = read_H263_horizontal_size;
					read_H263_ref_coded_picture_height = read_H263_coded_picture_height = read_H263_vertical_size;
					read_H263_ref_chrom_width = read_H263_chrom_width = read_H263_coded_picture_width >> 1;
					read_H263_ref_chrom_height = read_H263_chrom_height = read_H263_coded_picture_height >> 1;
					read_H263_source_format = read_H263_base_source_format;
				}
			}

			/* MMPTYPE */
			read_H263_pict_type = read_H263_getbits(3);
			if (read_H263_pict_type == PCT_IPB) read_H263_pb_frame = IM_PB_FRAMES;
			else read_H263_pb_frame = 0;

			if (PCT_B == read_H263_pict_type) {
				read_H263_true_B_frame = ON;
				/* Allow motion over picture boundaries, regardless of whether or
				 * not UMV is turned on. */
				read_H263_mv_outside_frame = 1;
				read_H263_true_b_trb = read_H263_temp_ref - read_H263_prev_non_disposable_temp_ref;
				if (read_H263_true_b_trb < 0) read_H263_true_b_trb += 256;
			} else {
				read_H263_true_B_frame = OFF;
				read_H263_prev_non_disposable_temp_ref = read_H263_next_non_disposable_temp_ref;
				read_H263_next_non_disposable_temp_ref = read_H263_temp_ref;
				read_H263_trd = read_H263_temp_ref - read_H263_prev_non_disposable_temp_ref;
				if (read_H263_trd < 0) read_H263_trd += 256;
			}

			read_H263_reference_picture_resampling_mode = read_H263_getbits(1);
			ASSERT(!read_H263_reference_picture_resampling_mode,(MSG("Reference picture resampling mode is not supported in this version")));
			read_H263_reduced_resolution_update_mode = read_H263_getbits(1);
			ASSERT(!read_H263_reduced_resolution_update_mode,(MSG("Reduced resolution update mode is not supported in this version")));
			read_H263_rtype = read_H263_getbits(1);      /* rounding type */
			tmp = read_H263_getbits(3);
			ASSERT(tmp == 1,(MSG("The last 3 bits of MPPTYPE is expected to be 001")));
		} else {
			/* UFEP is neither 001 nor 000 */
			ERROR(MSG("UFEP should be either 001 or 000."));
			END(1);
		}

		tmp = read_H263_getbits(1);
		ASSERT(!tmp,(MSG("CPM not supported in this version")));

		if (read_H263_UFEP && (read_H263_source_format == SF_CUSTOM)) {
			/* Read custom picture format */
			read_H263_CP_PAR_code = read_H263_getbits(4);
			ASSERT(read_H263_CP_PAR_code == PAR_CIF,(MSG("only 12:11 pixel aspect ratio supported")));
			tmp = read_H263_getbits(9);
			read_H263_horizontal_size = (tmp + 1 ) * 4;
			tmp = read_H263_getbits(1);
			ASSERT(tmp,(MSG("The 14th bit of Custom Picture Format(CPFMT) should be 1")));
			tmp = read_H263_getbits(9);
			read_H263_vertical_size = tmp * 4;
			ASSERT(!(read_H263_horizontal_size%16 || read_H263_vertical_size%16),(MSG("only factor of 16 custom source format supported")));

			if (read_H263_CP_PAR_code == EXTENDED_PAR) {
				read_H263_PAR_width = read_H263_getbits(8);
				read_H263_PAR_height = read_H263_getbits(8);
			}
		}

		if (read_H263_source_format != SF_CUSTOM) {
			read_H263_horizontal_size = read_H263_lines[read_H263_source_format];
			read_H263_vertical_size = read_H263_pels[read_H263_source_format];

		}
		read_H263_mb_width = read_H263_horizontal_size / 16;
		read_H263_mb_height = read_H263_vertical_size / 16;

		/* Need to store previous (reference layer) values
		 * for interpolation purposes, as the new values,
		 * i.e. of the spatially scaled layer, depend on
		 * the type of spatial scalability in use. */
		read_H263_ref_coded_picture_width = read_H263_coded_picture_width;
		read_H263_coded_picture_width = read_H263_horizontal_size;

		read_H263_ref_coded_picture_height = read_H263_coded_picture_height;
		read_H263_coded_picture_height = read_H263_vertical_size;

		read_H263_ref_chrom_width = read_H263_chrom_width;
		read_H263_chrom_width = read_H263_coded_picture_width >> 1;

		read_H263_ref_chrom_height = read_H263_chrom_height;
		read_H263_chrom_height = read_H263_coded_picture_height >> 1;

		if (read_H263_optional_custom_PCF) {
			if (read_H263_UFEP) {
				clock_conversion_code = read_H263_getbits(1);
				clock_divisor = read_H263_getbits(7);
				read_H263_CP_clock_frequency = (int) (1800 / ((float) clock_divisor * (8 + clock_conversion_code)) * 1000);
			}
			/* regardless of the value of UFEP */
			extended_temporal_reference = read_H263_getbits(2);
			read_H263_temp_ref = extended_temporal_reference << (8 + read_H263_temp_ref);

			if (PCT_B == read_H263_pict_type) {
				read_H263_true_b_trb = read_H263_temp_ref - read_H263_prev_non_disposable_temp_ref;
			} else {
				read_H263_trd = read_H263_temp_ref - read_H263_prev_non_disposable_temp_ref;
			}

			if (read_H263_trd < 0) read_H263_trd += 1024;
		}
		if (read_H263_UFEP && read_H263_long_vectors) {
			if (read_H263_getbits(1)) {
				read_H263_unlimited_unrestricted_motion_vectors = 0;
			} else {
				read_H263_flushbits(1);
				read_H263_unlimited_unrestricted_motion_vectors = 1;
			}
		}
		if (read_H263_UFEP && read_H263_slice_structured_mode) {
			read_H263_SSS_rectangular_slices = read_H263_getbits(1);
			read_H263_SSS_arbitary_slice_ordering = read_H263_getbits(1);
		}

		if ((read_H263_pict_type == PCT_B) || (read_H263_pict_type == PCT_EI) || (read_H263_pict_type == PCT_EP)) {
			/* optional temporal, SNR and spatial scalability mode in use */
			read_H263_enhancement_layer_num = read_H263_getbits(4);
			if (read_H263_UFEP) {
				read_H263_reference_layer_number = read_H263_getbits(4);
			}
			if (1 != read_H263_enhancement_layer_num) {
				if (read_H263_source_format != read_H263_base_source_format) {
					if (read_H263_source_format != SF_CUSTOM) {
						read_H263_scalability_mode = SPATIAL_SCALABILITY_HV;
					} else if (read_H263_coded_picture_width != read_H263_ref_coded_picture_width) {
						read_H263_scalability_mode = SPATIAL_SCALABILITY_H;
					} else {
						read_H263_scalability_mode = SPATIAL_SCALABILITY_V;
					}
				} else {
					read_H263_scalability_mode = SNR_SCALABILITY;
				}
			}
		} else {
			read_H263_enhancement_layer_num = read_H263_reference_layer_number = 1;
		}

		if (read_H263_reference_picture_selection_mode) {
			if (read_H263_UFEP) {
				read_H263_MF_of_reference_picture_selection = read_H263_getbits(3);
			}
			TRPI = read_H263_getbits(1);
			if (TRPI) {
				/* temporal reference for prediction exists */
				read_H263_temporal_reference_for_prediction = read_H263_getbits(10);
			}
			if (read_H263_showbits(1)) {
				/* backward channel message exists */
				/* BCM(backward channel message) is not implemented */
				read_H263_flushbits(1);
				ERROR(MSG("BCM(backward channel message) is not implemented in this version"));
				END(1);
			} else {
				/* BCM is not present */
				BCI = read_H263_getbits(2);
			}
		}
		ASSERT(!read_H263_reference_picture_resampling_mode,(MSG("RPRP reading is not implemented in this version")));
		read_H263_pic_quant = read_H263_getbits(5);
		read_H263_quant = read_H263_pic_quant;
	} else {
		read_H263_plus_type = 0;
		read_H263_rtype = 0;
		read_H263_source_format = tmp;
		read_H263_enhancement_layer_num = read_H263_reference_layer_number = 1;

		read_H263_horizontal_size = read_H263_lines[read_H263_source_format];
		read_H263_vertical_size = read_H263_pels[read_H263_source_format];

		read_H263_mb_width = read_H263_horizontal_size / 16;
		read_H263_mb_height = read_H263_vertical_size / 16;

		/* Need to store previous (reference layer) values
		 * for interpolation purposes, as the new values,
		 * i.e. of the spatially scaled layer, depend on
		 * the type of spatial scalability in use. */
		read_H263_ref_coded_picture_width = read_H263_coded_picture_width = read_H263_horizontal_size;
		read_H263_ref_coded_picture_height = read_H263_coded_picture_height = read_H263_vertical_size;
		read_H263_ref_chrom_width = read_H263_chrom_width = read_H263_coded_picture_width >> 1;
		read_H263_ref_chrom_height = read_H263_chrom_height = read_H263_coded_picture_height >> 1;

		read_H263_true_B_frame = OFF;
		read_H263_prev_non_disposable_temp_ref = read_H263_next_non_disposable_temp_ref;
		read_H263_next_non_disposable_temp_ref = read_H263_temp_ref;
		read_H263_trd = read_H263_temp_ref - read_H263_prev_non_disposable_temp_ref;
		if (read_H263_trd < 0) read_H263_trd += 256;

		read_H263_pict_type = read_H263_getbits(1);
		read_H263_mv_outside_frame = read_H263_getbits(1);
		read_H263_long_vectors = (read_H263_mv_outside_frame ? 1 : 0);
		read_H263_syntax_arith_coding = read_H263_getbits(1);
		read_H263_adv_pred_mode = read_H263_getbits(1);
		read_H263_mv_outside_frame = (read_H263_adv_pred_mode ? 1 : read_H263_mv_outside_frame);
		read_H263_overlapping_MC = (read_H263_adv_pred_mode ? 1 : 0);
		read_H263_use_4mv = (read_H263_adv_pred_mode ? 1 : 0);
		read_H263_pb_frame = read_H263_getbits(1);

		read_H263_pic_quant = read_H263_getbits(5);
		read_H263_quant = read_H263_pic_quant;

		tmp = read_H263_getbits(1);
		ASSERT(!tmp,(MSG("CPM not supported in this version")));
	}

	if (read_H263_pb_frame) {
		if (read_H263_optional_custom_PCF) {
			read_H263_trb = read_H263_getbits(5);
		} else {
			read_H263_trb = read_H263_getbits(3);
		}

		read_H263_bquant = read_H263_getbits(2);
	} else {
		read_H263_trb = 0;
	}

	pei = read_H263_getbits(1);

pspare:
	if (pei) {
		/* extra info for possible future backward compatible additions */
		read_H263_getbits(8);                /* not used */
		pei = read_H263_getbits(1);
		if (pei) goto pspare;              /* keep on reading pspare until pei=0 */
	}
}

/* get the GOB header */
/* author: <guyc@ece.ubc.ca */
void read_H263_getgobheader()
{
	int BCI = 0;
	/* GFID is not allowed to change unless PTYPE in picture header
	 * changes */
	read_H263_gfid = read_H263_getbits(2);
	/* NB: in error-prone environments the decoder can use this
	 * value to determine whether a picture header where the PTYPE
	 * has changed, has been lost */
	/* we now use it and assume that a picture change is *
	 * from INTRA to INTER or INTER to INTRA only */
	if (read_H263_gfid != read_H263_pgfid && read_H263_newgob) {
		if (read_H263_pict_type == PCT_INTRA) read_H263_pict_type = PCT_INTER;
		else if (read_H263_pict_type == PCT_INTER) read_H263_pict_type = PCT_INTRA;
	}

	read_H263_quant = read_H263_getbits(5);
	/* Support for Annex N at the GOB level */
	if (read_H263_reference_picture_selection_mode) {
		read_H263_TRI = read_H263_getbits(1);
		if (read_H263_TRI) {
			/* Temporal reference of GOB */
			read_H263_temp_ref = read_H263_optional_custom_PCF?read_H263_getbits(10):read_H263_getbits(8);
		}
		/* Temporal Reference for Prediction indication bit */
		read_H263_TRPI = read_H263_getbits(1);
		if (read_H263_TRPI) {
			/* temporal reference for prediction exists */
			read_H263_temporal_reference_for_prediction = read_H263_getbits(10);
		}
		if (read_H263_showbits(1)) {
			/* backward channel message exists */
			/* BCM(backward channel message) is not implemented */
			read_H263_flushbits(1);
			ERROR(MSG("BCM(backward channel message) is not implemented in this version"));
			END(1);
		} else {
			/* BCM is not present */
			BCI = read_H263_getbits(2);
		}
	}
	read_H263_pgfid = read_H263_gfid;
}

void read_H263_initoptions()
{
	read_H263_mv_outside_frame = 0;
	read_H263_syntax_arith_coding = 0;
	read_H263_adv_pred_mode = 0;
	read_H263_pb_frame = 0;
	read_H263_long_vectors = 0;

	read_H263_plus_type = 0;
	read_H263_optional_custom_PCF = 0;
	read_H263_advanced_intra_coding = 0;
	read_H263_deblocking_filter_mode = 0;
	read_H263_slice_structured_mode = 0;
	read_H263_reference_picture_selection_mode = 0;
	read_H263_independently_segmented_decoding_mode = 0;
	read_H263_alternative_inter_VLC_mode = 0;
	read_H263_modified_quantization_mode = 0;
	read_H263_reduced_resolution_update_mode = 0;
	read_H263_reference_picture_resampling_mode = 0;
	read_H263_rtype = 0;
}
